
<div class="slidecontainer">
  <input type="range" min="{{min_date}}" max="{{max_date}}" value="{{cur_date}}" class="slider" id="myRange" step="1">
  <div class="slidertext">{{name}}: <span id="demo"></span></div>
</div>